/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * Copyright 2007 Sun Microsystems, Inc. All rights reserved.
 */

package org.netbeans.server.uihandler.statistics;

import java.util.*;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;
import javax.servlet.jsp.PageContext;
import org.netbeans.server.uihandler.*;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import org.netbeans.server.uihandler.statistics.ParsingStatistics.Counts;
import org.openide.util.lookup.ServiceProvider;

/** Information about parsing length.
 *
 * @author Jaroslav Tulach
 */
@ServiceProvider(service=Statistics.class)
public class ParsingStatistics extends Statistics<ParsingStatistics.Counts> {
    static final Logger LOG = Logger.getLogger(ParsingStatistics.class.getName());
    /** minimal time considered idle */
    private static final long IDLE = 30000;
    
    public ParsingStatistics() {
        super("Parsing", 3);
    }
    
    ParsingStatistics(String n) {
        super(n);
    }
    
    protected Counts newData() {
        return null;
    }

    protected Counts process(LogRecord rec) {
        long when = rec.getMillis();
        if ("INDEXING_STARTED".equals(rec.getMessage())) { // NOI18N
            long time = Long.parseLong(rec.getParameters()[0].toString());
            return Counts.started(when, time);
        } else if ("INDEXING_FINISHED".equals(rec.getMessage())) { // NOI18N
            long time = Long.parseLong(rec.getParameters()[0].toString());
            return Counts.finished(when, time);
        }
        return Counts.createEvent(when);
    }

    protected Counts join(Counts one, Counts two) {
        return Counts.join(one, two);
    }
    
    @Override
    protected void registerPageContext(PageContext page, String name, Counts data) {
        LinkedHashMap<String,Long> map = new LinkedHashMap<String, Long>();
        if (data != null) {
            map.put("Indexing", data.indexing / 1000L); // NOI18N
            map.put("Ready", data.ready / 1000L); // NOI18N
            map.put("Idle", (data.idle + data.waiting) / 1000L); // NOI18N
        }
        page.setAttribute(name, map);
    }

    @Override
    protected Counts finishSessionUpload(String userId, int sessionNumber, boolean initialParse, Counts d) {
        return d;
    }

    public static final class Counts {
        final long indexing;
        final long ready;
        final long idle;
        final long waiting;
        transient final long when;
        
        Counts(long indexing, long ready, long activity, long idle, long waiting) {
            this.indexing = indexing;
            this.ready = ready;
            this.when = activity;
            this.idle = idle;
            this.waiting = waiting;
        }
        
        final boolean isTime() {
            return indexing == 0 && ready == 0 && idle == 0 && waiting == 0;
        }
        
        final boolean isParsing() {
            return indexing > 0 || waiting > 0;
        }
        
        public static Counts createEvent(long when) {
            return new Counts(0, 0, when, 0, 0);
        }
        
        public static Counts join(Counts one, Counts two) {
            if (one == null) {
                return two;
            }
            if (two == null) {
                return one;
            }
            final long when = Math.max(one.when, two.when);
            if (one.isTime() && two.isTime()) {
                return Counts.createEvent(when);
            }
            final long diff = Math.abs(one.when - two.when);
            Counts additional;
            if (diff > IDLE) {
                if (one.isParsing() || two.isParsing()) {
                    additional = Counts.waiting(when, diff);
                } else {
                    additional = Counts.idle(when, diff);
                }
            } else {
                additional = Counts.createEvent(when);
            }
            return new Counts(
                one.indexing + two.indexing + additional.indexing,
                one.ready + two.ready + additional.ready,
                when,
                one.idle + two.idle + additional.idle,
                one.waiting + two.waiting + additional.waiting
            );
        }

        @Override
        public String toString() {
            return "Counts{" + "indexing=" + indexing + 
                ", ready=" + ready + ", idle=" + idle + 
                ", waiting=" + waiting + ", when=" + when + '}';
        }
        
        public static Counts finished(long when, long time) {
            return new Counts(time, 0, when, 0, 0);
        }
        public static Counts started(long when, long time) {
            return new Counts(0, time, when, 0, 0);
        }
        public static Counts waiting(long when, long howLong) {
            return new Counts(0, 0, when, 0, howLong);
        }
        public static Counts idle(long when, long howLong) {
            return new Counts(0, 0, when, howLong, 0);
        }
    } // end of Counts
    
    protected void write(Preferences pref, Counts d) {
        if (d != null) {
            pref.putLong("indexing", d.indexing); // NOI18N
            pref.putLong("ready", d.ready); // NOI18N
            pref.putLong("idle", d.idle); // NOI18N
            pref.putLong("wait", d.waiting); // NOI18N
        }
    }

    protected Counts read(Preferences pref) throws BackingStoreException {
        final long indexing = pref.getLong("indexing", 0L); // NOI18N
        final long ready = pref.getLong("ready", 0L); // NOI18N
        final long idle = pref.getLong("idle", 0L); // NOI18N
        final long waiting = pref.getLong("wait", 0L); // NOI18N
        return indexing == 0L && idle == 0L ? null : new Counts(indexing, ready, -1, idle, waiting);
    }
}
