/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.cards.api;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.List;
import org.openide.util.NbBundle;

/**
 * A card game.  A Game provides String names for piles of cards which
 * may have different purposes.
 *
 * @author Tim Boudreau
 */
public interface Game {

    /**
     * Get the name of a pile.  In a multiplayer card game this might
     * be &quot;Joe's Hand&quot; - or whatever appropriate String
     * description for a particular pile of cards is appropriate,
     * like &quot;Discard Pile&quot;.
     * @param pile An instance of Pile
     * @return A localized name for a the Pile that was passed
     */
    public String nameOf(Pile pile);

    /**
     * Get the number of jokers a deck of cards for this game should
     * contain.  The return value from this method should always be
     * the same on repeated calls within the lifetime of one
     * Game object.
     * @return the number of jokers a deck to play this game should have
     */
    public int getJokerCount();

    /**
     * Default implementation of Game, as used by the Deck constructor
     * that takes an integer.  Just uses a cheap and cheezy accounting
     * technique to come up with names for piles by keeping a list of
     * all the ones it's been asked about in the past and assigning each
     * an integer.
     */
    static class DefaultGame implements Game {

        private List<Reference<Pile>> knownPiles = new ArrayList<Reference<Pile>>();
        private final int jokerCount;

        public DefaultGame(int jokerCount) {
            this.jokerCount = jokerCount;
        }

        public String nameOf(Pile pile) {
            int ix = 0;
            boolean found = false;
            for (Reference<Pile> ref : knownPiles) {
                if (pile == ref.get()) {
                    found = true;
                    break;
                }
                ix++;
            }
            if (!found) {
                knownPiles.add(new WeakReference<Pile>(pile));
                ix = knownPiles.size() - 1;
            }
            ix++;  //1 based indexing
            String result = NbBundle.getMessage(DefaultGame.class,
                    "FMT_PILE_NAME", new Integer(ix));
            return result;
        }

        public int getJokerCount() {
            return jokerCount;
        }
    }
}
