/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.cards.api;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Represents a pile of cards.  A Card from a given Deck may belong to
 * only one Pile at a time.  The pile a card is in determines whether
 * Card.getIdentity() and Card.getSuit() may legally be called.
 *
 * @author Tim Boudreau
 */
public final class Pile {

    private final List<Card> cards;
    private final AddPolicy addPolicy;

    public enum VisibilityPolicy {

        ALL_VISIBLE, TOP_CARD_VISIBLE, NO_CARDS_VISIBLE,
    }

    public Pile() {
        this(VisibilityPolicy.ALL_VISIBLE, AddPolicy.ALWAYS_ADD);
    }

    public Pile(VisibilityPolicy policy, AddPolicy addPolicy, List<Card> cards) {
        this(policy, addPolicy);
        this.cards.addAll(cards);
        for (Card card : cards) {
            card.pile = this;
        }
    }

    public Pile(VisibilityPolicy policy, AddPolicy addPolicy, Pile other) {
        this(policy, addPolicy, other.cards);
    }
    private final VisibilityPolicy policy;

    public Pile(VisibilityPolicy policy, AddPolicy addPolicy) {
        if (policy == null) {
            throw new NullPointerException("Null visibility policy");
        }
        if (addPolicy == null) {
            throw new NullPointerException("Null add policy");
        }
        this.policy = policy;
        this.addPolicy = addPolicy;
        cards = new ArrayList<Card>();
    }

    public Pile(VisibilityPolicy policy, List<Card> cards) {
        this(policy, AddPolicy.ALWAYS_ADD);
        this.cards.addAll(cards);
        cards = new ArrayList<Card>(cards);
        for (Card card : cards) {
            card.setPile(this);
        }
    }

    public void add(List<Card> cards) {
        for (Card card : cards) {
            add(card);
        }
    }

    public VisibilityPolicy getVisibilityPolicy() {
        return policy;
    }

    public boolean isEmpty() {
        return cards.isEmpty();
    }

    public Card peekTopCard() {
        return isEmpty() ? null : cards.get(0);
    }

    public Card peekBottomCard() {
        return isEmpty() ? null : cards.get(cards.size() - 1);
    }

    public Card removeTopCard() {
        Card card = cards.remove(0);
        fire();
        return card;
    }

    public Card removeBottomCard() {
        Card card = cards.remove(cards.size() - 1);
        fire();
        return card;
    }

    public List<Card> getCards() {
        return cards;
    }

    public Pile split() {
        return split(policy);
    }

    boolean isCardVisible(Card card) {
        if (card.pile() != this) {
            throw new IllegalArgumentException();
        }
        switch (policy) {
            case ALL_VISIBLE:
                return true;
            case TOP_CARD_VISIBLE:
                boolean result = card == cards.get(0);
                return result;
            case NO_CARDS_VISIBLE:
                return false;
            default:
                throw new AssertionError();
        }
    }

    public Pile split(VisibilityPolicy policy) {
        if (!isEmpty() && size() > 1) {
            List<Card> others = cards.subList(size() / 2, size());
            return new Pile(policy, others);
        } else {
            throw new IllegalStateException("Size is " + size());
        }
    }

    public int size() {
        return cards.size();
    }

    public Card cut() {
        Random r = new Random(System.currentTimeMillis());
        int ix = r.nextInt(size() - 1);
        return cards.get(ix);
    }

    public boolean canAdd(Card card) {
        return addPolicy.canAdd(card, this);
    }
    boolean inAdd;

    public void add(Card card) {
        if (inAdd) {
            return;
        }
        inAdd = true;
        try {
            if (canAdd(card)) {
                Pile old = card.getPile();
                if (old != null) {
                    card.getPile().remove(card);
                }
                cards.remove(card);
                cards.add(0, card);
                card.setPile(this);
                fire();
            } else {
                throw new IllegalArgumentException();
            }
        } finally {
            inAdd = false;
        }
    }

    public void remove(Card card) {
        if (!doRemove(card)) {
            throw new IllegalArgumentException(card + " not in " + this);
        }
    }

    boolean doRemove(Card card) {
        boolean result = cards.remove(card);
        fire();
        return result;
    }

    @Override
    public String toString() {
        return super.toString() + "[" + cards + "]";
    }
    private static List<ChangeListener> listeners =
            Collections.<ChangeListener>synchronizedList(new LinkedList<ChangeListener>());

    public void addChangeListener(ChangeListener l) {
        listeners.add(l);
    }

    public void removeChangeListener(ChangeListener l) {
        listeners.remove(l);
    }

    private void fire() {
        ChangeListener[] l = listeners.toArray(new ChangeListener[0]);
        for (int i = 0; i < l.length; i++) {
            l[i].stateChanged(new ChangeEvent(this));
        }
    }
}
