/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.nbtidyintegration;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.prefs.Preferences;
import java.util.regex.Matcher;
import org.openide.ErrorManager;
import org.openide.cookies.LineCookie;
import org.openide.cookies.SaveCookie;
import org.openide.execution.NbProcessDescriptor;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;
import org.openide.util.NbBundle;
import org.openide.windows.IOProvider;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;

/**
 * Do error checking on HTML files by calling tidy.
 *
 * Reads only error stream
 */
public class TidyRunnable implements Runnable {

    static final String EXECUTABLE_KEY = "tidyExecutable";
    private Node[] nodes;
    private String commandLineArgs;
    private String tidyExecutable = null;
    private boolean forceSave = false;

    /**
     * Initialize HTML-Tidy runnable.
     * @param nodes List of nodes to process
     * @param commandLineArgs Command line argument for the execution
     */
    public TidyRunnable(Node[] nodes, String commandLineArgs, boolean forceSave) {
        String defaultExecutablePath = NbBundle.getMessage(TidyErrorCheckAction.class, "AdvancedOption_Executable_Path_Default");
        tidyExecutable = Preferences.userNodeForPackage(
                TidyConfigurationPanel.class).get(EXECUTABLE_KEY, defaultExecutablePath);
        TidyAnnotation.clear();
        this.nodes = nodes;
        this.commandLineArgs = commandLineArgs;
        this.forceSave = forceSave;
    }

    public TidyRunnable(Node[] nodes, String commandLineArgs) {
        this(nodes, commandLineArgs, false);
    }

    /**
     * Perform error check, will be called in its own thread.
     */
    public void run() {
        InputOutput io = IOProvider.getDefault().getIO(
                NbBundle.getMessage(getClass(), "TidyOutputTabLabel"), false);
        io.select(); //Tree tab is selected
        OutputWriter writer = io.getOut();

        try {
            writer.reset(); //clean the output window
            writer.println("Selected " + nodes.length + " objects ...");

            if (new File(getTidyExecutable()).exists()) {// Check, if executable exits
                for (int i = 0; i < nodes.length; i++) {
                    DataObject dataObject = (DataObject) nodes[i].getCookie(DataObject.class);
                    processDataObject(writer, getTidyExecutable(), dataObject);
                }
            } else {
                writer.println("Executable not found at " + getTidyExecutable() + ". Please select a valid path in the options panel.");
            }

        } catch (Exception ex) {
            ErrorManager.getDefault().notify(ErrorManager.WARNING, ex);
        }
        writer.close();
    }

    /**
     * Perform external program with this data object.
     * @param executable Absolute path to the executable
     * @param writer Writer for the output tab
     * @param dataObject Data Object that is processed
     * @throws java.lang.InterruptedException
     * @throws java.io.IOException
     */
    void processDataObject(final OutputWriter writer,
            final String executable, final DataObject dataObject)
            throws InterruptedException, IOException {
        if (forceSave) {
            forceSave(dataObject);
        }
        FileObject fileObject = dataObject.getPrimaryFile();
        File file = FileUtil.toFile(fileObject);
        writer.println("File: " + file + " - Size: " + file.length() + "...");

        NbProcessDescriptor tidyProcessDesc =
                new NbProcessDescriptor(executable,
                getCommandLineArgs() + " \"" + file.getAbsolutePath() + "\"");
        Process process = tidyProcessDesc.exec();
        readOutput(writer, dataObject, process.getErrorStream());
        process.waitFor();
        writer.println("Exit: " + process.exitValue());
        writer.flush();
    }

    /**
     * Read the process output and error stream simultaneously. Return the
     * content of the output stream in a string. Print the error
     * messages to the screen and generate annotations if the error string
     * matches a pattern.
     * @param writer Writer for the output tab
     * @param dataObject Data Object that is processed
     * @param errStream Process error stream
     */
    static void readOutput(OutputWriter writer, DataObject dataObject,
            InputStream errStream) {
        LineCookie lc = (LineCookie) dataObject.getCookie(LineCookie.class);
        TidyOutputListener listener = new TidyOutputListener(dataObject);
        try {
            BufferedReader error = new BufferedReader(new InputStreamReader(errStream));
            String err = null;
            while ((err = error.readLine()) != null) {
                Matcher matcher = TidyOutputListener.PATTERN.matcher(err);

                if (matcher.matches()) {
                    TidyAnnotation.createAnnotation(lc, matcher);
                    writer.println(err, listener);
                } else {
                    writer.println(err);
                }
            }
        } catch (Exception e) {
            ErrorManager.getDefault().notify(ErrorManager.WARNING, e);
        }
    }

    public String getCommandLineArgs() {
        return commandLineArgs;
    }

    public String getTidyExecutable() {
        return tidyExecutable;
    }

    public void setTidyExecutable(String tidyExecutable) {
        this.tidyExecutable = tidyExecutable;
    }

    /** Save the Data Object, if it has been modified*/
    void forceSave(DataObject dataObject) throws IOException {
        if (dataObject.isModified()) {
            SaveCookie cookie = (SaveCookie) dataObject.getCookie(SaveCookie.class);
            if (cookie != null) {
                cookie.save();
            }
        }
    }
}
