/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.customexplorer.filetype;

import java.awt.datatransfer.Transferable;
import java.awt.dnd.DnDConstants;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.Collections;
import java.util.List;
import org.openide.loaders.DataNode;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.nodes.NodeAdapter;
import org.openide.nodes.NodeEvent;
import org.openide.nodes.NodeTransfer;
import org.openide.util.Lookup;
import org.openide.util.datatransfer.PasteType;
import org.netbeans.examples.customexplorer.CustomRectangle;
import org.netbeans.examples.customexplorer.ShapesBusinessObject;
import org.netbeans.examples.customexplorer.ShapesContentListener;

public class ShapesDataNode extends DataNode {

    private static final String IMAGE_ICON_BASE = "org/netbeans/examples/" +
            "customexplorer/filetype/shapes.png";

    public ShapesDataNode(final ShapesDataObject obj) {
        super(obj, new ShapesChildren(obj));
        setIconBaseWithExtension(IMAGE_ICON_BASE);
    }

    ShapesDataNode(final ShapesDataObject obj, final Lookup lookup) {
        super(obj, new ShapesChildren(obj), lookup);
        setIconBaseWithExtension(IMAGE_ICON_BASE);
    }

    @Override
    protected void createPasteTypes(Transferable t, List<PasteType> s) {
        super.createPasteTypes(t, s);
        PasteType paste = getDropType(t, DnDConstants.ACTION_COPY, -1);
        if (null != paste) {
            System.err.println("Created a paste type " + paste);
            s.add(paste);
        }
    }

    @Override
    public PasteType getDropType(Transferable t, final int action, int index) {
        final Node dropNode = NodeTransfer.node(t,
                DnDConstants.ACTION_COPY_OR_MOVE + NodeTransfer.CLIPBOARD_CUT);

        if (dropNode != null) {
            final CustomRectangle rect =
                    dropNode.getLookup().lookup(CustomRectangle.class);

            if (rect != null && !this.equals(dropNode.getParentNode())) {
                return new PasteTypeImpl(rect);
            }
        }

        return null;
    }

    private static class ShapesChildren extends Children.Keys {
        private ShapesDataObject dataObj;

        private NodeReaperAdapter reaper;

        ShapesChildren(ShapesDataObject dataObj) {
            this.dataObj = dataObj;

            reaper = new NodeReaperAdapter();

            // recalculate children when shapes added or removed
            ShapesBusinessObject bizObj = dataObj.getShapesBusinessObject();
            bizObj.addShapesContentListener(new ShapesContentListener() {
                public void shapesAdded() {
                    refreshList();
                }

                public void shapesRemoved() {
                    refreshList();
                }
            });
        }


        @Override
        protected void addNotify() {
            refreshList();
        }

        @Override
        @SuppressWarnings("unchecked")
        protected void removeNotify() {
            setKeys(Collections.EMPTY_SET);
        }

        @Override
        protected Node[] createNodes(Object key) {
            CustomRectangle rect = (CustomRectangle) key;
            RectangleNode rectangleNode = new RectangleNode(rect);

            // TODO hook up bi-directional node handling as in TDL
            rect.addPropertyChangeListener(new PropertyChangeListener() {
                public void propertyChange(PropertyChangeEvent evt) {
                    dataObj.setModified(true);
                }
            });

            // so model gets updated when node is destroyed
            // we could override RectangleNode.destroy instead, but we do not have
            // acess to the ShapesBusinessObject over there
            rectangleNode.addNodeListener(reaper);

            return new Node[] { rectangleNode };
        }

        @SuppressWarnings("unchecked")
        private void refreshList() {
            ShapesDataNode node = (ShapesDataNode)dataObj.getNodeDelegate();
            ShapesBusinessObject bizObj = node.getLookup().lookup(
                    ShapesBusinessObject.class);
            if (bizObj == null) {
                System.err.println("ERROR: ShapesBusinessObject in lookup was null");
                return;
            }

            setKeys(bizObj.getAllRectangles());
        }

        private class NodeReaperAdapter extends NodeAdapter {
            @Override
            public void nodeDestroyed(NodeEvent evt) {
                System.out.println("node was removed");
                Node corpse = evt.getNode();
                corpse.removeNodeListener(this); // allow GC

                CustomRectangle cr = corpse.getLookup().lookup(
                        CustomRectangle.class);

                ShapesBusinessObject sbo = dataObj.getShapesBusinessObject();
                sbo.remove(cr);
            }
        }
    }

    private class PasteTypeImpl extends PasteType {
        private final CustomRectangle rect;
        private PasteTypeImpl(CustomRectangle rect) {
            this.rect = rect;
        }

        @Override
        public Transferable paste() throws IOException {
            ShapesDataObject targetGdo = getLookup().lookup(ShapesDataObject.class);
            targetGdo.getShapesBusinessObject().add(rect);
            return null;
        }
    }
}