/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 * 
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.qa.reports;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;

/**
 * Set of often used queries.
 *
 * @author Jan Stola
 */
public class Queries {
    /** Verbose mode? */
    public static boolean verbose=false;
    /** Cache of loaded Issues. */
    static Map cache=new HashMap();
    
    /**
     * Returns numbers of all open issues in given <code>module</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param module the module that contains the issues.
     * @throws IOException
     */
    public static List getIssueNumbers(String module, Set versions) throws IOException {
        return getIssueNumbers(module, 0, versions);
    }
    
    /**
     * Returns numbers of all open issues in given <code>modules</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param modules the modules that contains the issues.
     * @throws IOException
     */
    public static List getIssueNumbers(Set modules, Set versions) throws IOException {
        return getIssueNumbers(modules, 0, versions);
    }
    
    /**
     * Returns numbers of all open issues in given <code>module</code> with specified
     * <code>issueType</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param issueType
     * @param versions set of versions
     * @param module the module that contains the issues.
     * @throws IOException
     */
    public static List getIssueNumbers(String module, int issueType, Set versions) throws IOException {
        Set modules=new HashSet(1);
        modules.add(module);
        return getIssueNumbers(modules, issueType, versions);
    }
    
    /**
     * Returns numbers of all open issues in given <code>modules</code> with specified
     * <code>issueType</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param issueType
     * @param versions set of versions
     * @param modules the modules that contains the issues.
     * @throws IOException
     */
    public static List getIssueNumbers(Set modules, int issueType, Set versions) throws IOException {
        IssueQuery iq=new IssueQuery();
        iq.type=issueType;
        iq.product=modules;
        iq.version=versions;
        iq.status=Issue.STATUS_OPEN;
        return iq.getIssueNumbers();
    }
    
    /**
     * Returns numbers of all open defects in given <code>module</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param module the module that contains the defects.
     * @throws IOException
     */
    public static List getBugNumbers(String module, Set versions) throws IOException {
        return getIssueNumbers(module, Issue.TYPE_DEFECT, versions);
    }
    
     /**
      * Returns numbers of all open defects in given <code>modules</code>.
      *
      * @return <code>List</code> of <code>Integer</code>s.
      * @param versions set of versions
      * @param modules the modules that contains the defects.
      * @throws IOException
      */
    public static List getBugNumbers(Set modules, Set versions) throws IOException {
        return getIssueNumbers(modules, Issue.TYPE_DEFECT, versions);
    }
    
    /**
     * Returns numbers of all open defects in given <code>module</code>
     * with specified <code>keyword</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param module the module that contains the defects.
     * @param keyword the keyword that the defect should include.
     * @throws IOException
     */
    public static List getBugNumbers(String module, String keyword, Set versions) throws IOException {
        Set modules=new HashSet(1);
        modules.add(module);
        return getBugNumbers(modules, keyword, versions);
    }
    
    /**
     * Returns numbers of all open defects in given <code>modules</code>
     * with specified <code>keyword</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param modules the modules that contains the defects.
     * @param keyword the keyword that the defect should include.
     * @throws IOException
     */
    public static List getBugNumbers(Set modules, String keyword, Set versions) throws IOException {
        IssueQuery iq=new IssueQuery();
        iq.type=Issue.TYPE_DEFECT;
        iq.product=modules;
        iq.version=versions;
        iq.status=Issue.STATUS_OPEN;
        Set keywords=new HashSet(1);
        keywords.add(keyword);
        iq.keywords=keywords;
        return iq.getIssueNumbers();
    }
    
    /**
     * Returns numbers of all open defects in given <code>module</code>
     * with specified <code>priority</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param module the module that contains the defects.
     * @param priority mask for the priority of the bugs.
     * @throws IOException
     */
    public static List getBugNumbers(String module, int priority, Set versions) throws IOException {
        Set modules=new HashSet(1);
        modules.add(module);
        return getBugNumbers(modules, priority, versions);
    }
    
    /**
     * Returns numbers of all open defects in given <code>modules</code>
     * with specified <code>priority</code>.
     *
     * @return <code>SortedSet</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param modules the modules that contains the defects.
     * @param priority mask for the priority of the bugs.
     * @throws IOException
     */
    public static List getBugNumbers(Set modules, int priority, Set versions) throws IOException {
        IssueQuery iq=new IssueQuery();
        iq.type=Issue.TYPE_DEFECT;
        iq.product=modules;
        iq.version=versions;
        iq.status=Issue.STATUS_OPEN;
        iq.priority=priority;
        return iq.getIssueNumbers();
    }
    
    /**
     * Returns numbers of all open enhancements in given <code>module</code>.
     *
     * @return <code>SortedSet</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param module the module that contains the enhancements.
     * @throws IOException
     */
    public static List getEnhancementNumbers(String module, Set versions) throws IOException {
        return getIssueNumbers(module, Issue.TYPE_ENHANCEMENT, versions);
    }
    
    /**
     * Returns numbers of all open enhancements in given <code>modules</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param modules the modules that contains the enhancements.
     * @throws IOException
     */
    public static List getEnhancementNumbers(Set modules, Set versions) throws IOException {
        return getIssueNumbers(modules, Issue.TYPE_ENHANCEMENT, versions);
    }
    
    /**
     * Returns numbers of all open features in given <code>module</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param module the module that contains the features.
     * @throws IOException
     */
    public static List getFeatureNumbers(String module, Set versions) throws IOException {
        return getIssueNumbers(module, Issue.TYPE_FEATURE, versions);
    }
    
    /**
     * Returns numbers of all open features in given <code>modules</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param modules the modules that contains the features.
     * @throws IOException
     */
    public static List getFeatureNumbers(Set modules, Set versions) throws IOException {
        return getIssueNumbers(modules, Issue.TYPE_FEATURE, versions);
    }
    
    /**
     * Returns numbers of all open tasks in given <code>module</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param module the module that contains the tasks.
     * @throws IOException
     */
    public static List getTaskNumbers(String module, Set versions) throws IOException {
        return getIssueNumbers(module, Issue.TYPE_TASK, versions);
    }
    
    /**
     * Returns numbers of all open tasks in given <code>modules</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param modules the modules that contains the tasks.
     * @throws IOException
     */
    public static List getTaskNumbers(Set modules, Set versions) throws IOException {
        return getIssueNumbers(modules, Issue.TYPE_TASK, versions);
    }
    
    /**
     * Returns numbers of all open patches in given <code>module</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param module the module that contains the patches.
     * @throws IOException
     */
    public static List getPatchNumbers(String module, Set versions) throws IOException {
        return getIssueNumbers(module, Issue.TYPE_PATCH, versions);
    }
    
    /**
     * Returns numbers of all open patches in given <code>modules</code>.
     *
     * @return <code>List</code> of <code>Integer</code>s.
     * @param versions set of versions
     * @param modules the modules that contains the patches.
     * @throws IOException
     */
    public static List getPatchNumbers(Set modules, Set versions) throws IOException {
        return getIssueNumbers(modules, Issue.TYPE_PATCH, versions);
    }
    
    /**
     * Returns counts of open issues in given <code>module</code>.
     *
     * @param versions set of versions
     * @param module the module for which the distribution of issues is generated.
     * @throws IOException
     * @return
     */
    public static IssueCounts getIssueCounts(String module, Set versions) throws IOException {
        IssueQuery iq=new IssueQuery();
        Set modules=new HashSet(1);
        modules.add(module);
        iq.product=modules;
        iq.version=versions;
        iq.status=Issue.STATUS_OPEN;
        return iq.getIssueCounts();
    }
    
    /**
     * Returns counts of open bugs in given <code>module</code>.
     *
     * @param versions set of versions
     * @param module the module for which the distribution of issues is generated.
     * @throws IOException
     * @return
     */
    public static IssueCounts getBugCounts(String module, Set versions) throws IOException {
        IssueQuery iq=new IssueQuery();
        Set modules=new HashSet(1);
        modules.add(module);
        iq.type=Issue.TYPE_DEFECT;
        iq.product=modules;
        iq.version=versions;
        iq.status=Issue.STATUS_OPEN;
        return iq.getIssueCounts();
    }
    
     /**
      * Returns issue for the given <code>issueNumber</code>.
      *
      * @return <code>Issue</code> specified by the <code>issueNumber</code>.
      * @param issueNumber desired issue.
      * @throws IOException  */
    public static Issue getIssue(int issueNumber) throws IOException {
        List issueNumbers=new LinkedList();
        issueNumbers.add(new Integer(issueNumber));
        List issues=getIssues(issueNumbers);
        return (Issue)issues.get(0);
    }
    
    /**
     * Returns issues for the given list of <code>issueNumbers</code>.
     *
     * @param issueNumbers numbers of the desired issues.
     * @return <code>List</code> of <code>Issue</code>s.
     */
    public static List getIssues(List issueNumbers) throws IOException {
        boolean runParser=false;
        StringBuffer sb=new StringBuffer("http://www.netbeans.org/issues/xml.cgi?id=");
        List cached=new LinkedList();
        Iterator iter=issueNumbers.iterator();
        while (iter.hasNext()) {
            Object number=iter.next();
            Object issue=cache.get(number);
            if (issue!=null) {
                cached.add(issue);
            } else {
                sb.append(number).append("%2C");
                runParser=true;
            }
        }
        if (runParser) {
            URL url=new URL(sb.substring(0,sb.length()-3).toString());
            HttpURLConnection con=(HttpURLConnection)url.openConnection();
            XMLExportHandler handler = new XMLExportHandler();
            SAXParserFactory factory = SAXParserFactory.newInstance();
            try {
                SAXParser saxParser=factory.newSAXParser();
                saxParser.parse(con.getInputStream(), handler);
            } catch (IOException e) {
                throw e;
            } catch (Exception e) {
                e.printStackTrace();
            }
            cached.addAll(handler.getIssues());
            iter=handler.getIssues().iterator();
            while (iter.hasNext()) {
                Issue issue=(Issue)iter.next();
                cache.put(new Integer(issue.number), issue);
            }
        }
        SortedSet sorted=new TreeSet(new IssueComparator());
        sorted.addAll(cached);
        cached=new LinkedList(sorted);
        return cached;
    }
    
    /**
     * Comparator for issues. Major criterium is issue priority,
     * minor criterium is issue number.
     */
    public static class IssueComparator implements Comparator {
        public int compare(Object o1, Object o2) {
            Issue i1=(Issue)o1;
            Issue i2=(Issue)o2;
            return ((i1.priority-i2.priority)!=0)? (i1.priority-i2.priority) : (i1.number-i2.number);
        }
    }
    
    /**
     * Returns numbers of issues changed since given <code>date</code>.
     *
     * @param date reference date.
     * @return <code>List</code> of <code>Integer</code>s.
     */
    public static List getChangedIssueNumbers(Date date) throws IOException {
        List issueNumbers=new LinkedList();
        SimpleDateFormat sdf=new SimpleDateFormat("yyyy-MM-dd%20HH:mm:ss");
        URL url=new URL("http://www.netbeans.org/issues/updated_since.cgi?ts="+sdf.format(date));
        //System.out.println("Changed issues URL: " + url.toExternalForm());
        HttpURLConnection con=(HttpURLConnection)url.openConnection();
        try {
            BufferedReader br=new BufferedReader(new InputStreamReader(con.getInputStream()));
            try {
                String s;
                while (((s=br.readLine())!=null)&&(!s.equals("<PRE>")));
                boolean end=false;
                while (((s=br.readLine())!=null)&&(!end)) {
                    if (s.endsWith("</PRE>")) {
                        s=s.substring(0, s.length()-6);
                        end=true;
                    }
                    if (s.trim().equals("")) break;
                    try {                        
                        issueNumbers.add(new Integer(s));
                    } catch (NumberFormatException e) {
                        e.printStackTrace();
                    }
                }
            } finally {
                br.close();
            }
        } finally {
            con.disconnect();
        }
        return issueNumbers;
    }
    
    /**
     * Returns activity of issue specified by <code>issueNumber</code>.
     *
     * @param issueNumber number of the desired issue.
     * @return activity of issue specified by <code>issueNumber</code>.
     */
    public static IssueActivity getActivity(int issueNumber) throws IOException {
        if (verbose) System.out.println("Activity for issue "+issueNumber+" obtained.");
        IssueActivity activity=new IssueActivity();
        URL url=new URL("http://www.netbeans.org/issues/show_activity.cgi?id="+issueNumber);
        HttpURLConnection con=(HttpURLConnection)url.openConnection();
        try {
            BufferedReader br=new BufferedReader(new InputStreamReader(con.getInputStream()));
            try {
                String s;
                while (((s=br.readLine())!=null)&&(!s.equals("    <th>Who</th><th>What</th><th>Old value</th><th>New value</th><th>When</th>")));
                br.readLine();
                while (((s=br.readLine())!=null)&&(s.equals("<tr>"))) {
                    Change change=new Change();
                    s=br.readLine(); s=s.substring(4,s.length()-5);
                    change.who=s;
                    s=br.readLine(); s=s.substring(4,s.length()-5);
                    change.what=s;
                    s=br.readLine(); s=s.substring(4,s.length()-5);
                    change.oldValue=s;
                    s=br.readLine(); s=s.substring(4,s.length()-5);
                    change.newValue=s;
                    s=br.readLine(); s=s.substring(4,s.length()-5);
                    SimpleDateFormat sdf=new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
                    try {
                        change.when=sdf.parse(s);
                    } catch (ParseException e) {
                        e.printStackTrace();
                    }
                    br.readLine();
                    activity.add(change);
                }
            } finally {
                br.close();
            }
        } finally {
            con.disconnect();
        }
        return activity;
    }
    
}
